////////////////////////////////////////////////////////////////////////////
//  
//  KRepton - the classic Repton game for KDE
//  
//  Copyright (c) 1998 Sandro Sigala <sandro@sigala.it>
//  Copyright (c) 2003 Jonathan Marten <jjm@keelhaul.demon.co.uk>
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//  
//  It is distributed in the hope that it will be useful, but WITHOUT ANY
//  WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING for further details.
//  If not, write to the Free Software Foundation, Inc., 59 Temple Place -
//  Suite 330, Boston, MA 02111-1307, USA.
//
////////////////////////////////////////////////////////////////////////////

#include "config.h"

#include <kapp.h>
#include <kmessagebox.h>
#include <kglobal.h>
#include <kstandarddirs.h>

#include <qbitmap.h>
#include <qfile.h>
#include <qpainter.h>
#include <qpixmap.h>

#include "krepton.h"
#include "gamewidget.h"
#include "map.h"

GameWidget::GameWidget(QWidget *parent,const char *name)
        : QWidget(parent,name)
{
	in_game = false;
	in_pause = false;
	sprites = 0;
	maps = 0;

	setFocusPolicy(QWidget::StrongFocus);
	setFocus();

	timerObjects = startTimer(200);
	timerMonsters = startTimer(300);
	timerEggs = startTimer(500);
	timerSecs = startTimer(1000);
	timerPlants = startTimer(3000);

	bgpix = new QPixmap(KGlobal::dirs()->findResource("graphics","bg.bmp"));
	pausepix = new QPixmap(KGlobal::dirs()->findResource("graphics","pause.bmp"));
}

GameWidget::~GameWidget()
{
	delete bgpix;
	delete pausepix;
}

void GameWidget::loadEpisode(const QString& path)
{
	loadSprites(path);
	loadMaps(path);
}

void GameWidget::loadSprites(const QString& path)
{
	QPixmap allsprites;

	allsprites.load(path + "/sprites.bmp");
	if (sprites) delete [] sprites;
	sprites = new QPixmap[num_sprites];
// TODO: change 6 and 8 to constants (see editor)
	for (int y = 0; y<6; ++y)
	{
		for (int x = 0; x<8; ++x)
		{
			int i = y*8 + x; 
			sprites[i].resize(sprite_width,sprite_height);
			QPainter p;
			p.begin(&sprites[i]);
			p.drawPixmap(0,0,allsprites,
				     x*sprite_width,y*sprite_height,
				     sprite_width,sprite_height);
			p.end();
		}
	}

	sprites[Obj::Blip].setMask(sprites[Obj::Blip].createHeuristicMask());
	sprites[Obj::Blip2].setMask(sprites[Obj::Blip2].createHeuristicMask());
}

void GameWidget::loadMaps(const QString& path)
{
	QFile f(path + "/info");
	if (!f.open(IO_ReadOnly)) fatalExit(QString("Cannot read 'info' file in '%1'").arg(path));

	QTextStream t(&f);
	QString name;
	name = t.readLine();
	t >> num_levels;
	f.close();

	if (num_levels<1) fatalExit(QString("Episode at '%1' has no levels").arg(path));

	if (maps) delete [] maps;
	maps = new Map[num_levels];
	for (int i = 0; i<num_levels; ++i)
	{
		const QString p = QString("%1/map%2").arg(path).arg(i+1);
		maps[i].loadData(p);
	}
}

void GameWidget::startGame(const QString& episode, int level, int points)
{
	episode_name = episode;

	current_map = &maps[level];
	current_level = level;
	current_map->setPoints(points);

	QString msg = QString("<qt>Starting level %1 of episode <b>%2</b>")
		.arg(level+1).arg(episode);
	if (level>0) msg += QString("<br>The password is \"<b>%1</b>\"")
			     .arg(current_map->getPassword());
	KMessageBox::information(this,msg);

	in_game = true;
	in_pause = false;
	emit changedState(in_game,in_pause);
	current_map->startGame();
}

void GameWidget::startGamePassword(const QString& episode, const QString& password)
{
	for (int i = 0; i < num_levels; ++i)
	{
		if (maps[i].getPassword().lower()==password.lower())
		{
			startGame(episode,i,0);
			return;
		}
	}
	KMessageBox::sorry(this,"Wrong password");
}

void GameWidget::timerEvent(QTimerEvent *e)
{
	if (!in_game || in_pause) return;

	if (current_map->hasDied() || current_map->hasEndedLevel())
	{
		in_game = false;
		emit changedState(in_game,in_pause);
		if (current_map->hasDied()) endedGame(current_map->howDied());
		else
		{
			if ((current_level+1)<num_levels)
			{
				KMessageBox::information(this,"Level finished!");
				startGame(episode_name,current_level+1,
					  current_map->getPoints());
			}
			else
			{
				KMessageBox::information(this,"Episode finished!");
				emit gameOver();
			}
		}

		repaint(false);
		return;
	}

	int id = e->timerId();

	if (id==timerObjects)
	{
		current_map->updateObjects();
		repaint(false);
	}
	else if (id==timerMonsters)
	{
		current_map->updateMonsters();
		repaint(false);
	}
	else if (id==timerEggs)
	{
		current_map->updateEggs();
		current_map->updateObjects();
		repaint(false);
	}
	else if (id==timerSecs)
	{
		if (--current_map->num_secs<=0) endedGame("Your time is up!");
	}
	else if (id==timerPlants)
	{
		current_map->updatePlants();
		repaint(false);
	}

	emit changedStats(current_map->num_diamonds,current_map->num_secs,
			  current_map->num_points);
	emit changedFlags(current_map->have_key,current_map->have_crown);
}

void GameWidget::paintEvent(QPaintEvent *)
{
	if (!in_game)
	{
		for (int y = 0; y<=(height()/bgpix->height()); ++y)
		{
			for (int x = 0; x<=(width()/bgpix->width()); ++x)
			{
				bitBlt(this,x*bgpix->width(),y*bgpix->height(),bgpix);
			}
		}
		return;
	}

	QPixmap pm(size());
	QPainter p;
	p.begin(&pm);
	for (int y = 0; y<=(height()/sprite_height); y++)
	{
		for (int x = 0; x<=(width()/sprite_width); x++)
		{
			int mx = x + current_map->xpos - (width()/sprite_width/2);
			int my = y + current_map->ypos - (height()/sprite_height/2);

			SpriteObj obj = current_map->xy(mx,my);
			switch (obj)
			{
case Obj::Broken_Egg1:
case Obj::Broken_Egg2:
case Obj::Broken_Egg3:		obj = Obj::Broken_Egg;
				break;

case Obj::Falling_Egg:		obj = Obj::Egg;
				break;

case Obj::Repton:		obj = current_map->sprite;
				break;

default:			;			// Avoid warning
			}

			Monster *m;
			if ((m = current_map->findMonster(mx, my)) != 0
			    && m->type == Obj::Monster)
				obj = m->sprite;

			p.drawPixmap(x * sprite_width, y * sprite_height, sprites[obj]);

			if (m != 0 && m->type == Obj::Blip)
				p.drawPixmap(x * sprite_width, y * sprite_height, sprites[m->sprite]);
		}
	}

	if (in_pause) p.drawPixmap((width()-pausepix->width())/2, 
				   (height()-pausepix->height())/2, *pausepix);
	p.end();
	bitBlt(this,0,0,&pm);
}

void GameWidget::goUp()
{
	if (!in_game || in_pause) return;
	current_map->goUp();
	repaint(false);
}

void GameWidget::goDown()
{
	if (!in_game || in_pause) return;
	current_map->goDown();
	repaint(false);
}

void GameWidget::goLeft()
{
	if (!in_game || in_pause) return;
	current_map->goLeft();
	repaint(false);
}

void GameWidget::goRight()
{
	if (!in_game || in_pause) return;
	current_map->goRight();
	repaint(false);
}

void GameWidget::keyPressEvent(QKeyEvent *e)
{
	switch (e->key())
	{
case Key_Up:	goUp();
		break;

case Key_Down:	goDown();
		break;

case Key_Left:	goLeft();
		break;

case Key_Right:	goRight();
		break;

default:	e->ignore();
		break;
	}
}

void GameWidget::endedGame(const QString &how)
{
//	kdDebug(0) << "GameWidget::endedGame" << endl;

	in_game = false;
	in_pause = false;
	emit changedState(in_game,in_pause);

	KMessageBox::sorry(this,QString("You died...\n%1").arg(how));
	emit gameOver();
}


void GameWidget::pauseAction()
{
//	kdDebug(0) << "GameWidget::pauseAction()" << endl;

	if (!in_game) return;

	in_pause = !in_pause;
	emit changedState(in_game,in_pause);
	repaint(false);
}

void GameWidget::endAction()
{
//	kdDebug(0) << "GameWidget::endAction()" << endl;

	if (!in_game) return;
	endedGame("You killed yourself!");
}
