////////////////////////////////////////////////////////////////////////////
//  
//  KRepton - the classic Repton game for KDE
//  
//  Copyright (c) 1998 Sandro Sigala <sandro@sigala.it>
//  Copyright (c) 2003 Jonathan Marten <jjm@keelhaul.demon.co.uk>
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//  
//  It is distributed in the hope that it will be useful, but WITHOUT ANY
//  WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING for further details.
//  If not, write to the Free Software Foundation, Inc., 59 Temple Place -
//  Suite 330, Boston, MA 02111-1307, USA.
//
////////////////////////////////////////////////////////////////////////////

#include "config.h"

#include <kapplication.h>
#include <kmessagebox.h>
#include <kstatusbar.h>
#include <kmainwindow.h>
#include <kconfig.h>
#include <kglobal.h>
#include <kstandarddirs.h>
#include <kstdaction.h>
#include <kstdaccel.h>
#include <kaction.h>
#include <kpassdlg.h>

#include <qbitmap.h>

#include "krepton.h"
#include "gamewindow.h"
#include "gamewidget.h"
#include "scoredialog.h"
#include "newscoredialog.h"
#include "splashwindow.h"

static const int minimum_width = 5*sprite_width;
static const int minimum_height = 3*sprite_height;
static const int default_width = 9*sprite_width;
static const int default_height = 9*sprite_height;

bool GameWindow::do_sounds = true;

GameWindow::GameWindow(QWidget *parent,const char *name)
        : KMainWindow(parent,name)
{
	kdDebug(0) << "GameWindow::GameWindow name='" << name << "'" << endl;
	setCaption(name);

	(void) new KAction("&Quit",KStdAccel::shortcut(KStdAccel::Quit),
			   this,SLOT(slotQuit()),
			   actionCollection(),"game_quit");

	newGameList = new KSelectAction("&New",0 /* KStdAccel::shortcut(KStdAccel::New) */,
					this,SLOT(slotNewGame()),
					actionCollection(),"game_new");

	contGameList = new KSelectAction("&Continue",0 /* CTRL+Key_C */,
					 this,SLOT(slotContinueGame()),
					 actionCollection(),"game_continue");

	pauseAction = new KToggleAction("&Pause",Key_P,
					this,SLOT(slotPauseGame()),
					actionCollection(),"game_pause");

	endGameAction = new KAction("&Finish",Key_Escape,
				    this,SLOT(slotEndGame()),
				    actionCollection(),"game_end");

	(void) new KAction("&High Scores...",CTRL+Key_H,
			   this,SLOT(slotHighScores()),
			   actionCollection(),"game_highscores");

	soundsAction = new KToggleAction("&Sounds",CTRL+Key_S,
					 this,SLOT(slotToggleSounds()),
					 actionCollection(),"settings_sounds");

	loadSpritesList = new KSelectAction("&Load Sprites",0,
					    this,SLOT(slotLoadSprites()),
					    actionCollection(),"settings_loadsprites");

	updateState(false,false);
	createGUI();

	loadEpisodeList(episodes);			// global and local episodes

	QStringList list;
	for (Episode *e = episodes.first(); e != 0; e = episodes.next()) list.append(e->name);
	newGameList->setItems(list);
	contGameList->setItems(list);
	loadSpritesList->setItems(list);

	QPixmap *keypix = new QPixmap(KGlobal::dirs()->findResource("graphics","key.png"));
	const QSize ss = keypix->size();
	QPixmap *crownpix = new QPixmap(KGlobal::dirs()->findResource("graphics","crown.png"));
	keypix->setMask(keypix->createHeuristicMask());
	crownpix->setMask(crownpix->createHeuristicMask());

	KStatusBar *status = statusBar();
	status->insertFixedItem("Diamonds: 999",1);
	status->insertFixedItem("Time: 9999",2);
	status->insertFixedItem("Score: 999999",3);

	keyflag = new QLabel(this,QString::null,Qt::WStyle_Customize|Qt::WStyle_NoBorder);
	keyflag->setFixedSize(ss);
	keyflag->setPixmap(*keypix);
	keyflag->setEnabled(false);

	crownflag = new QLabel(this,QString::null,Qt::WStyle_Customize|Qt::WStyle_NoBorder);
	crownflag->setFixedSize(ss);
	crownflag->setPixmap(*crownpix);
	crownflag->setEnabled(false);

	status->addWidget(keyflag,0,true);
	status->addWidget(crownflag,0,true);

	for (int i = 1; i<=3; ++i) status->setItemAlignment(i,Qt::AlignLeft|Qt::AlignVCenter);
	updateStats(-1);

	game = new GameWidget(this);
	setCentralWidget(game);

	connect(game, SIGNAL(changedStats(int,int,int)),
		this, SLOT(updateStats(int,int,int)));
	connect(game, SIGNAL(changedFlags(bool,bool)),
		this, SLOT(updateFlags(bool,bool)));

	connect(game,SIGNAL(changedState(bool,bool)),this,SLOT(updateState(bool,bool)));
	connect(game,SIGNAL(gameOver()),this,SLOT(gameOver()));

	toolBar("mainToolBar")->hide();

	QSize s1(minimum_width,minimum_height);
//	game->setMinimumSize(s1);
	setBaseSize(sizeForCentralWidgetSize(s1));
	setSizeIncrement(sprite_width,sprite_height);
	QSize s2(default_width,default_height);
	resize(sizeForCentralWidgetSize(s2));

	setAutoSaveSettings();
	readOptions();

	finalizeGUI(false);
	updateState(false,false);

	SplashWindow *splash = new SplashWindow();
	splash->show();
	kdDebug(0) << "GameWindow::GameWindow done" << endl;
}

void GameWindow::readOptions()
{
	kdDebug(0) << "GameWindow::readOptions" << endl;

	KConfig *config = KGlobal::config();
	config->setGroup("Options");

	soundsAction->setChecked(config->readBoolEntry("Sounds",true));
	slotToggleSounds();
}

void GameWindow::saveOptions()
{
	kdDebug(0) << "GameWindow::saveOptions" << endl;

	KConfig *config = KGlobal::config();
	config->setGroup("Options");
	config->writeEntry("Sounds",soundsAction->isChecked());
}

bool GameWindow::queryClose()
{
	saveOptions();
	return (true);
}

void GameWindow::slotQuit()
{
	close();
}

void GameWindow::slotHighScores()
{
	ScoreDialog d(episodes,this);
	d.exec();
}

void GameWindow::slotToggleSounds()
{
	do_sounds = soundsAction->isChecked();
}

void GameWindow::updateStats(int diamonds,int secs,int points)
{
	KStatusBar *status = statusBar();

	status->changeItem((diamonds<0 ? QString::null : QString("Diamonds: %1").arg(diamonds)),1);
	status->changeItem((secs<0 ? QString::null : QString("Time: %1").arg(secs)),2);
	status->changeItem((points<0 ? QString::null : QString("Score: %1").arg(points)),3);
}

void GameWindow::updateFlags(bool key,bool crown)
{
	keyflag->setEnabled(key);
	crownflag->setEnabled(crown);
}

void GameWindow::updateState(bool ingame,bool inpause)
{
	pauseAction->setChecked(inpause);
	pauseAction->setEnabled(ingame);
	endGameAction->setEnabled(ingame);

	loadSpritesList->setEnabled(ingame);
	newGameList->setEnabled(!ingame);
	contGameList->setEnabled(!ingame);
}

void GameWindow::gameOver()
{
	updateFlags(false,false);
	updateStats(-1);

	KConfig *config = KGlobal::config();
	config->setGroup("High Score");
	QString s, score;
	score = config->readEntry(current_episode->filename + "Score", "0");
	if (game->getPoints()>score.toInt())
	{
		NewScoreDialog d(this);
		if (d.exec())
		{
			score.setNum(game->getPoints());
			config->writeEntry(current_episode->filename + "Score", score);
			config->writeEntry(current_episode->filename + "Name", d.name());

			ScoreDialog h(episodes,this);
			h.exec();
		}
	}
}

QString GameWindow::prepareGame(int sel,bool spritesonly)
{
	if (sel==-1) return (QString::null);
	Episode *e = episodes.at(sel);
	if (!spritesonly) current_episode = e;

	QString path = KGlobal::dirs()->findResourceDir("episodes",e->filename+"/info");
	kdDebug(0) << "GameWindow::prepareGame name=" << e->name << " file=" << e->filename << " path=" << path << endl;

	if (path==QString::null)
	{
		KMessageBox::error(this,QString("Cannot find episode files for '%1'.\nIs the application properly installed?").arg(e->filename));
		return (QString::null);
	}

	loadSpritesList->setCurrentItem(sel);
	if (!spritesonly) game->loadEpisode(path+e->filename);
	else game->loadSprites(path+e->filename);
	return (e->name);
}

void GameWindow::slotNewGame()
{
	const int selection = newGameList->currentItem();
	newGameList->setCurrentItem(-1);

//	kdDebug(0) << "GameWindow::slotNewGame selected=" << selection << endl;

	QString name = prepareGame(selection,false);
	if (name==QString::null) return;
	game->startGame(name,0);
}

void GameWindow::slotContinueGame()
{
	const int selection = contGameList->currentItem();
	contGameList->setCurrentItem(-1);

//	kdDebug(0) << "GameWindow::slotContinueGame selected=" << selection << endl;

	QString name = prepareGame(selection,false);
	if (name==QString::null) return;

	QCString pwd;
	if (KPasswordDialog::getPassword(pwd,QString("<qt>Continuing episode <b>%1</b>").arg(name))!=KPasswordDialog::Accepted) return;
	pwd = pwd.stripWhiteSpace();
	if (pwd.isEmpty()) KMessageBox::information(this,"No idea? You need to start at the first level.");
	else game->startGamePassword(name,pwd);
}

void GameWindow::slotLoadSprites()
{
	const int selection = loadSpritesList->currentItem();

//	kdDebug(0) << "GameWindow::slotLoadSprites selected=" << selection << endl;

	prepareGame(selection,true);
	game->repaint(false);
}

void GameWindow::slotPauseGame()
{
	kdDebug(0) << "GameWindow::slotPauseGame "<< endl;
	game->pauseAction();
}

void GameWindow::slotEndGame()
{
	kdDebug(0) << "GameWindow::slotEndGame "<< endl;
	game->endAction();
}
