////////////////////////////////////////////////////////////////////////////
//  
//  KRepton - the classic Repton game for KDE
//  
//  Copyright (c) 1998 Sandro Sigala <sandro@sigala.it>
//  Copyright (c) 2003 Jonathan Marten <jjm@keelhaul.demon.co.uk>
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//  
//  It is distributed in the hope that it will be useful, but WITHOUT ANY
//  WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
//  for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING for further details.
//  If not, write to the Free Software Foundation, Inc., 59 Temple Place -
//  Suite 330, Boston, MA 02111-1307, USA.
//
////////////////////////////////////////////////////////////////////////////

#include "config.h"

#include <kapp.h>
#include <kglobal.h>
#include <kstandarddirs.h>
#include <kaudioplayer.h>

#include <qfile.h>
#include <qtstream.h>

#include <stdlib.h>
#include <math.h>

#include "krepton.h"
#include "gamewindow.h"
#include "map.h"

Map::Map()
{
	monsters.setAutoDelete(true);
	transporters.setAutoDelete(true);
	how_died = QString::null;
	status_end_level = false;

	sounddir = KGlobal::dirs()->findResourceDir("sound","die.wav");
}

Map::~Map()
{
	transporters.clear();
	monsters.clear();
	delete[] data;
}

void Map::loadData(const QString& filename)
{
	QFile f(filename);
	if (!f.open(IO_ReadOnly)) fatalExit(QString("Cannot read map file '%1'").arg(filename));

	QTextStream t(&f);
	int num_transporters;
	t >> width >> height >> num_transporters >> num_secs;
	t >> password;

	while (num_transporters--)
	{
		int ox, oy, dx, dy;
		t >> ox >> oy >> dx >> dy;
		addTransporter(ox, oy, dx, dy);
	}

	data = new SpriteObj[width * height];

	num_diamonds = num_points = 0;
	have_key = have_crown = false;
	sprite = Obj::Repton;

	bool got_position = false;
	for (int y = 0; y<height; ++y)
	{
		char c;
		for (int x = 0; x<width; ++x)
		{
			t >> c;
			ref(x,y) = (SpriteObj)(c-'A');
			switch (ref(x, y))
			{
case Obj::Blip:			ref(x,y) = Obj::Empty;
				addMonster(x,y,Obj::Blip);
				break;

case Obj::Strongbox:
case Obj::Cage:
case Obj::Diamond:		++num_diamonds;
				break;

case Obj::Plant:			addMonster(x,y,Obj::Plant);
				break;

case Obj::Repton:		if (got_position) fatalExit(QString("Inconsistent map: duplicate starting position"));
				got_position = true;
				xpos = x;
				ypos = y;
				break;

default:			;			// Avoid warning
			}
		}
	}

	if (!got_position) fatalExit(QString("Inconsistent map: no starting position"));

	f.close();
	prepareMonsters();
}

void Map::playSound(Sound::Type s)
{
	if (!GameWindow::soundsEnabled()) return;

	QString name;
	switch (s)
	{
case Sound::Die:
		name = "die";
		break;

case Sound::Got_Diamond:
		name = "diamond";
		break;

case Sound::Got_Crown:
		name = "crown";
		break;

case Sound::Cage_Blip:
		name = "cage";
		break;

case Sound::Got_Key:
		name = "key";
		break;

case Sound::Got_Time:
		name = "time";
		break;

case Sound::Transport:
		name = "transport";
		break;

case Sound::Kill_Monster:
		name = "monster";
		break;

case Sound::Broken_Egg:
		name = "egg";
		break;
	}

	QString fname = sounddir+name+".wav";
	KAudioPlayer::play(fname);
}

// This makes sure that KNotify is running at the start of the game.
// Otherwise, there may be a big delay when the first diamond is hit...
void Map::startGame()
{
	playSound(Sound::Transport);
}

void Map::addTransporter(int ox, int oy, int dx, int dy)
{
	transporters.append(new Transporter(ox, oy, dx, dy));
}

Transporter *Map::findTransporter(int x, int y)
{
	for (Transporter *t = transporters.first(); t!=0; t = transporters.next())
	{
		if (t->orig_x==x && t->orig_y==y) return (t);
	}

	fatalExit(QString("Inconsistent map: invalid transporter"));
	return (0);
}

void Map::addMonster(int x, int y, SpriteObj type)
{
	monsters.append(new Monster(x,y,type));
}

void Map::prepareMonsters()
{
	for (Monster *m = monsters.first(); m != 0; m = monsters.next())
	{
		if (m->type!=Obj::Blip) continue;

		const int x = m->xpos;
		const int y = m->ypos;

		if (!isempty(x-1,y) && isempty(x,y-1)) m->orientation = Orientation::North;
		else if (!isempty(x,y-1) && isempty(x+1,y)) m->orientation = Orientation::East;	
		else if (!isempty(x+1,y) && isempty(x,y+1)) m->orientation = Orientation::South;
		else if (!isempty(x,y+1) && isempty(x-1,y)) m->orientation = Orientation::West;
	}
}

Monster *Map::findMonster(int x, int y)
{
	for (Monster *m = monsters.first(); m != 0; m = monsters.next())
	{
		if (m->xpos==x && m->ypos==y) return (m);
	}
	return (0);
}

// Try falling down.
bool Map::tryFallDown(int x, int y)
{
	SpriteObj obj = xy(x,y);
	const SpriteObj next1 = xy(x,y+1);
	const SpriteObj next2 = xy(x,y+2);

	if (next1!=Obj::Empty) return (false);
	ref(x,y) = Obj::Empty;

	Monster *m;
	if ((m = findMonster(x,y+1))!=0 && m->type==Obj::Monster)
	{
		playSound(Sound::Kill_Monster);
		killMonster(m);
	}

	if (obj==Obj::Rock && (next2==Obj::Broken_Egg || next2==Obj::Broken_Egg1 ||
			       next2==Obj::Broken_Egg2 || next2==Obj::Broken_Egg3))
	{
		playSound(Sound::Kill_Monster);
		ref(x,y+2) = Obj::Empty;
	}

// TODO: falling egg can land on other things too?
	if (next2==Obj::Wall && (obj==Obj::Egg || obj==Obj::Falling_Egg))
	{
		playSound(Sound::Broken_Egg);
		obj = Obj::Broken_Egg;
	}

	if (next2==Obj::Repton) die("You got crushed!");

	if (obj==Obj::Egg) obj = Obj::Falling_Egg;
	ref(x,y+1) = obj;
	return (true);
}

// Check if the specified object can fall to the left or to the right.
bool Map::tryFallHorizontal(int x, int y, int xd)
{
	const SpriteObj obj = xy(x,y);

	if (xy(x+xd,y+1)==Obj::Empty && xy(x+xd,y)==Obj::Empty)
	{
		ref(x,y) = Obj::Empty;
		if (obj==Obj::Egg) ref(x+xd,y) = Obj::Falling_Egg;
		else ref(x+xd,y) = obj;
		return (true);
	}

	return (false);
}

void Map::tryFallLeftOrRight(int x, int y)
{
	const SpriteObj obj = xy(x,y);

	switch(xy(x,y+1))
	{
case Obj::Rock:
case Obj::Skull:
case Obj::Diamond:
case Obj::Egg:
case Obj::Key:
case Obj::Bomb:
case Obj::Broken_Egg:
case Obj::Broken_Egg1:
case Obj::Broken_Egg2:
case Obj::Broken_Egg3:
		// Curved objects.  Rocks and eggs can fall onto these objects.
		// If can't fall to the left, fall to the right.
		if (!tryFallHorizontal(x,y,-1))
		{
			if (!tryFallHorizontal(x,y,1) && obj==Obj::Falling_Egg)
			{
				playSound(Sound::Broken_Egg);
				ref(x,y) = Obj::Broken_Egg;
			}
		}
		break;

case Obj::Wall_North_West:
case Obj::Filled_Wall_North_West:
		if (!tryFallHorizontal(x,y,-1) && obj==Obj::Falling_Egg)
		{
			playSound(Sound::Broken_Egg);
			ref(x,y) = Obj::Broken_Egg;
		}
		break;		

case Obj::Wall_North_East:
case Obj::Filled_Wall_North_East:
		if (!tryFallHorizontal(x,y,1) && obj==Obj::Falling_Egg)
		{
			playSound(Sound::Broken_Egg);
			ref(x,y) = Obj::Broken_Egg;
		}
		break;

default:	if (obj==Obj::Falling_Egg)
		{
			playSound(Sound::Broken_Egg);
			ref(x,y) = Obj::Broken_Egg;
		}
	}
}

// Check if the specified object can fall.
void Map::tryFall(int x, int y)
{
	if (!tryFallDown(x, y))
		tryFallLeftOrRight(x, y);
}

// Check if the blip can go into a direction.
bool Map::blipTryDirection(Monster *m,Orientation::Type dir,int xd,int yd)
{
	switch (xy(m->xpos+xd,m->ypos+yd))
	{
case Obj::Empty:
case Obj::Ground1:
case Obj::Ground2:
		m->xpos += xd;
		m->ypos += yd;
		m->orientation = dir;
		return (true);

case Obj::Cage:
		cageBlip(m, m->xpos + xd, m->ypos + yd);
		return (true);

case Obj::Repton:
		die("It got you!");
		break;

default:	;					// Avoid warning
	}

	return (false);
}

// Check and move the blip.
void Map::updateBlip(Monster *m)
{
	const int mx = m->xpos;
	const int my = m->ypos;

	switch (m->orientation)
	{
case Orientation::North:
		if (!blipTryDirection(m, Orientation::West,-1,0) &&
		    !blipTryDirection(m, Orientation::North,0,-1))
		{
			m->orientation = Orientation::East;
			updateBlip(m);
			return;
		}
		break;

case Orientation::South:
		if (!blipTryDirection(m, Orientation::East,1,0) &&
		    !blipTryDirection(m, Orientation::South,0,1))
		{
			m->orientation = Orientation::West;
			updateBlip(m);
			return;
		}
		break;

case Orientation::East:
		if (!blipTryDirection(m, Orientation::North,0,-1) &&
		    !blipTryDirection(m, Orientation::East,1,0))
		{
			m->orientation = Orientation::South;
			updateBlip(m);
			return;
		}
		break;

case Orientation::West:
		if (!blipTryDirection(m, Orientation::South,0,1) &&
		    !blipTryDirection(m, Orientation::West,-1,0))
		{
			m->orientation = Orientation::North;
			updateBlip(m);
			return;
		}
		break;
	}

	if (m->xpos != mx || m->ypos != my)
		m->sprite = (m->sprite==Obj::Blip) ? Obj::Blip2 : Obj::Blip;
}

// Check if the monster can go into a direction, and calculate the
// distance between it and Repton.
double Map::monsterTryDirection(Monster *m, int xd, int yd)
{
	switch (xy(m->xpos+xd, m->ypos+yd))
	{
case Obj::Empty:
case Obj::Ground1:
case Obj::Ground2:
		return (sqrt(pow(m->xpos+xd-xpos,2) + pow(m->ypos+yd-ypos,2)));

case Obj::Repton:
		die("The monster got you!");

default:	;
	}

	return (666);					// A rather large number
}

// Check the Repton position and move the monster according with it.
void Map::updateMonster(Monster *m)
{
	int xd = 0, yd = 0;
	double i, len;

	len = sqrt(pow(m->xpos-xpos,2) + pow(m->ypos-ypos,2));

	if ((i = monsterTryDirection(m, 0, -1))<len)
	{
		xd = 0; yd = -1; len = i;
	}
	if ((i = monsterTryDirection(m, 0, 1))<len)
	{
		xd = 0; yd = 1; len = i;
	}
	if ((i = monsterTryDirection(m, 1, 0))<len)
	{
		xd = 1; yd = 0; len = i;
	}
	if (monsterTryDirection(m, -1, 0)<len)
	{
		xd = -1; yd = 0;
	}

	m->xpos += xd;
	m->ypos += yd;
	m->sprite = (m->sprite==Obj::Monster) ? Obj::Monster2 : Obj::Monster;
}

// Check the broken eggs.
void Map::updateEggs()
{
	for (int y = height - 1; y >= 0; y--)
	{
		for (int x = width - 1; x >= 0; x--)
		{
			switch (xy(x,y))
			{
case Obj::Broken_Egg:		ref(x,y) = Obj::Broken_Egg1;
				break;

case Obj::Broken_Egg1:		ref(x,y) = Obj::Broken_Egg2;
				break;

case Obj::Broken_Egg2:		ref(x,y) = Obj::Broken_Egg3;
				break;

case Obj::Broken_Egg3:		ref(x,y) = Obj::Empty;
				addMonster(x,y,Obj::Monster);
				break;

default:			;			// Avoid warning
			}
		}
	}
}

// Check the moveable objects (rocks and eggs).
void Map::updateObjects()
{
	for (int y = height-2; y>=0; y--)		// Note the ``height-2'' value.
	{
		for (int x = width - 1; x >= 0; x--)
		{
			switch(xy(x,y))
			{
case Obj::Rock:
case Obj::Egg:
case Obj::Falling_Egg:
case Obj::Broken_Egg:
case Obj::Broken_Egg1:
case Obj::Broken_Egg2:
case Obj::Broken_Egg3:		tryFall(x,y);		// Check if the object can fall.
				break;

default:			;			// Avoid warning
			}
		}
	}
}

// Try to reproduce the plant.
bool Map::tryPlant(int x, int y, int dx, int dy)
{
	if (xy(x+dx,y+dy)==Obj::Repton)
	{
		die("The plant got you!");
		return (true);
	}

	if (xy(x+dx,y+dy)==Obj::Empty)
	{
		ref(x+dx,y+dy) = Obj::Plant;
		addMonster(x+dx,y+dy,Obj::Plant);
		return (true);
	}

	return (false);
}

// Check the plants and reproduce.
bool Map::updatePlant(Monster *m)
{
	if ((rand() % 100)<=10)
	{
		if (tryPlant(m->xpos, m->ypos, 1, 0)) return (true);
		else if (tryPlant(m->xpos, m->ypos, -1, 0)) return (true);
		else if (tryPlant(m->xpos, m->ypos, 0, 1)) return (true);
		else if (tryPlant(m->xpos, m->ypos, 0, -1)) return (true);
	}
	return (false);
}

// Reproduce the plants
void Map::updatePlants()
{
	for (Monster *m = monsters.first(); m!=0; m = monsters.next())
	{
		if (m->type==Obj::Plant)
		{
			if (updatePlant(m)) break;
		}
	}
}

// Check the monsters status and move them.
void Map::updateMonsters()
{
	for (Monster *m = monsters.first(); m!=0; m = monsters.next())
	{
		switch (m->type)
		{
case Obj::Blip:		updateBlip(m);
			break;

case Obj::Monster:	updateMonster(m);
			break;

default:		;				// Avoid warning
		}
	}
}

// Move Repton vertically.
void Map::moveVerticalDirect(int yd)
{
	// Move Repton up or down according with the yd value.
	ref(xpos, ypos) = Obj::Empty;
	ref(xpos, ypos + yd) = Obj::Repton;
	ypos += yd;

	// Change the Repton sprite orientation.
	switch (sprite)
	{
case Obj::Repton:
case Obj::Repton_Look_Left:
case Obj::Repton_Look_Right:
case Obj::Repton_Go_Right1:
case Obj::Repton_Go_Right2:
case Obj::Repton_Go_Right3:
case Obj::Repton_Go_Right4:
case Obj::Repton_Go_Left1:
case Obj::Repton_Go_Left2:
case Obj::Repton_Go_Left3:
case Obj::Repton_Go_Left4:
case Obj::Repton_Go_Up2:
		sprite = Obj::Repton_Go_Up1;
		break;

case Obj::Repton_Go_Up1:
		sprite = Obj::Repton_Go_Up2;
		break;

default:	;					// Avoid warning
	}
}

// Move Repton horizontally.
void Map::moveHorizontalDirect(int xd)
{
	// Move Repton to the left or to the right according
	// with the xd value.
	ref(xpos, ypos) = Obj::Empty;
	ref(xpos + xd, ypos) = Obj::Repton;
	xpos += xd;

	// Change the Repton sprite orientation.
	if (xd==-1)					// Move Repton to the left.
	{
		switch (sprite)
		{
case Obj::Repton:
case Obj::Repton_Look_Left:
case Obj::Repton_Look_Right:
case Obj::Repton_Go_Up1:
case Obj::Repton_Go_Up2:
case Obj::Repton_Go_Right1:
case Obj::Repton_Go_Right2:
case Obj::Repton_Go_Right3:
case Obj::Repton_Go_Right4:
case Obj::Repton_Go_Left4:
			sprite = Obj::Repton_Go_Left1;
			break;

case Obj::Repton_Go_Left1:
			sprite = Obj::Repton_Go_Left2;
			break;

case Obj::Repton_Go_Left2:
			sprite = Obj::Repton_Go_Left3;
			break;

case Obj::Repton_Go_Left3:
			sprite = Obj::Repton_Go_Left4;
			break;

default:		;				// Avoid warning
		}
	}
	else						// Move Repton to the right.
	{
		switch (sprite)
		{
case Obj::Repton:
case Obj::Repton_Look_Left:
case Obj::Repton_Look_Right:
case Obj::Repton_Go_Up1:
case Obj::Repton_Go_Up2:
case Obj::Repton_Go_Left1:
case Obj::Repton_Go_Left2:
case Obj::Repton_Go_Left3:
case Obj::Repton_Go_Left4:
case Obj::Repton_Go_Right4:
			sprite = Obj::Repton_Go_Right1;
			break;

case Obj::Repton_Go_Right1:
			sprite = Obj::Repton_Go_Right2;
			break;

case Obj::Repton_Go_Right2:
			sprite = Obj::Repton_Go_Right3;
			break;

case Obj::Repton_Go_Right3:
			sprite = Obj::Repton_Go_Right4;
			break;

default:		;				// Avoid warning
		}
	}
}

// This function is called when a key is taken. It makes the strongboxes into diamonds.
void Map::gotKey()
{
	if (have_key) return;				// Already have the key

	for (int y = 0; y<height; ++y)
	{
		for (int x = 0; x<width; ++x)
		{
			if (xy(x,y)==Obj::Strongbox) ref(x,y) = Obj::Diamond;
		}
	}
	have_key = true;
}

// This function is called when Repton get an object.
void Map::gotObject(SpriteObj obj)
{
	switch (obj)
	{
case Obj::Diamond:
		playSound(Sound::Got_Diamond);
		--num_diamonds;
		num_points += 50;
		break;

case Obj::Time:	playSound(Sound::Got_Time);
		num_secs += 30;
		num_points += 10;
		break;

case Obj::Key:	playSound(Sound::Got_Key);
		gotKey();
		num_points += 30;
		break;

case Obj::Crown:
		playSound(Sound::Got_Crown);
		num_points += 200;
		have_crown = true;
		break;

case Obj::Bomb:	if (num_diamonds==0) status_end_level = true;
		break;

default:	;					// Avoid warning
	}
}

// Move Repton to a new map location when it enters a transport.
void Map::useTransporter(int x, int y)
{
	Transporter *t = findTransporter(x, y);

	// Remove the transporter and Repton from the original location.
	ref(x, y) = Obj::Empty;
	ref(xpos, ypos) = Obj::Empty;

	playSound(Sound::Transport);

	xpos = t->dest_x;
	ypos = t->dest_y;
	switch (xy(xpos,ypos))
	{
case Obj::Diamond:
case Obj::Time:
case Obj::Key:
case Obj::Crown:
		gotObject(xy(xpos, ypos));
		// Move Repton to the new location.
		ref(xpos, ypos) = Obj::Repton;
		sprite = Obj::Repton;
		break;

case Obj::Ground1:
case Obj::Ground2:
		num_points += 2;
		/* FALLTHROUGH */

case Obj::Empty:
		if (findMonster(xpos, ypos) != 0) die("You transported onto a monster!");
		// Move Repton to the new location.
		ref(xpos, ypos) = Obj::Repton;
		sprite = Obj::Repton;
		break;

case Obj::Transport:
		useTransporter(xpos,ypos);
		break;

default:	die("You transported onto something!");
		break;
	}
}

// Check if Repton can move vertically according with the yd value.
void Map::moveVertical(int yd)
{
	SpriteObj obj = xy(xpos, ypos + yd);

	if (findMonster(xpos, ypos + yd) != 0) die("You ran into a monster!");

	switch (obj)
	{
case Obj::Diamond:
case Obj::Time:
case Obj::Crown:
case Obj::Key:	gotObject(obj);
		/* FALLTHROUGH */

case Obj::Empty:
		moveVerticalDirect(yd);
		break;

case Obj::Ground1:
case Obj::Ground2:
		num_points += 2;
		moveVerticalDirect(yd);
		break;

case Obj::Transport:
		useTransporter(xpos, ypos + yd);
		break;

case Obj::Skull:
case Obj::Plant:
		die("You ran into something fatal!");
		break;

case Obj::Bomb:	gotObject(Obj::Bomb);
		break;

default:	;					// Avoid warning
	}
}

// Move Repton and a near object horizontally.
void Map::moveHorizontalMoveObj(int xd, SpriteObj obj)
{
	if (xy(xpos + xd*2, ypos)==Obj::Empty)
	{
		Monster *m;
		if ((m = findMonster(xpos + xd*2, ypos))!=0 && m->type==Obj::Monster)
		{
			playSound(Sound::Kill_Monster);
			killMonster(m);
		}
		ref(xpos + xd*2, ypos) = obj;
		moveHorizontalDirect(xd);
	}
}

// Check if Repton can move horizontally according with the xd value.
void Map::moveHorizontal(int xd)
{
	SpriteObj obj = xy(xpos + xd, ypos);

	if (findMonster(xpos + xd, ypos) != 0) die("You ran into a monster!");

	switch (obj)
	{
case Obj::Rock:
case Obj::Egg:
case Obj::Broken_Egg:
case Obj::Broken_Egg1:
case Obj::Broken_Egg2:
case Obj::Broken_Egg3:
		moveHorizontalMoveObj(xd, obj);
		break;

case Obj::Diamond:
case Obj::Time:
case Obj::Crown:
case Obj::Key:	gotObject(obj);
		/* FALLTHROUGH */

case Obj::Empty:
		moveHorizontalDirect(xd);
		break;

case Obj::Ground1:
case Obj::Ground2:
		num_points += 2;
		moveHorizontalDirect(xd);
		break;

case Obj::Transport:
		useTransporter(xpos + xd, ypos);
		break;

case Obj::Skull:
case Obj::Plant:
		die("You ran into something fatal!");
		break;

case Obj::Bomb:
		gotObject(Obj::Bomb);
		break;

default:	;					// Avoid warning
	}
}

// Kill a monster.
void Map::killMonster(Monster *mp)
{
	num_points += 500;
	for (Monster *m = monsters.first(); m!=0; m = monsters.next())
	{
		if (m==mp)
		{
			monsters.remove();
			break;
		}
	}
}

// Kill a blip, make it a diamond, and produce a sound.
void Map::cageBlip(Monster *m, int x, int y)
{
	playSound(Sound::Cage_Blip);
	ref(x,y) = Obj::Diamond;
	killMonster(m);
}

void Map::die(const QString &how)
{
	kdDebug(0) << "Map::die()" << endl;

	how_died = how;
	playSound(Sound::Die);
}
